import logging
import platform
from pathlib import Path

import pluggy
from ocrmypdf import OcrEngine, hookimpl
from ocrmypdf._exec import tesseract
from ocrmypdf.exceptions import ExitCodeException
from PIL import Image

from ocrmypdf_appleocr.common import Textbox, lang_code_to_locale, log
from ocrmypdf_appleocr.hocr import build_hocr_document
from ocrmypdf_appleocr.livetext import (
    livetext_supported,
    ocr_VKCImageAnalyzerRequest,
    supported_languages_livetext,
)
from ocrmypdf_appleocr.pdf import generate_pdf
from ocrmypdf_appleocr.vision import (
    ocr_VNRecognizeTextRequest,
    supported_languages_accurate,
    supported_languages_fast,
)

__version__ = "0.3.1"


def perform_ocr(image: Path, options) -> tuple[list[Textbox], int, int, tuple[int, int]]:
    im = Image.open(image)
    width, height = im.size
    dpi = im.info["dpi"]

    if options.appleocr_recognition_mode == "livetext":
        locales = [lang_code_to_locale.get(lang, lang) for lang in options.languages]
        textboxes = ocr_VKCImageAnalyzerRequest(image, width, height, locales)
    else:
        textboxes = ocr_VNRecognizeTextRequest(image, width, height, options)

    return textboxes, width, height, dpi


@hookimpl
def initialize(plugin_manager: pluggy.PluginManager):
    # Disable built-in Tesseract OCR engine to avoid conflict
    plugin_manager.set_blocked("ocrmypdf.builtin_plugins.tesseract_ocr")


@hookimpl
def add_options(parser):
    appleocr_options = parser.add_argument_group("Apple OCR", "Apple Vision OCR options")
    appleocr_options.add_argument(
        "--appleocr-disable-correction",
        action="store_true",
        help="Disable language correction in Apple Vision OCR (default: False)",
        default=False,
    )
    appleocr_options.add_argument(
        "--appleocr-recognition-mode",
        choices=["fast", "accurate", "livetext"],
        default="livetext" if livetext_supported else "accurate",
        help="Recognition mode for Apple Vision OCR (default: accurate for macOS 12 and earlier, livetext for macOS 13 and later)",
    )


@hookimpl
def check_options(options):
    if options.languages:
        if len(options.languages) == 1 and options.languages[0] == "und":
            if options.appleocr_recognition_mode == "livetext":
                raise ExitCodeException(
                    15, "Language detection is not supported by LiveText mode in Apple OCR"
                )
            options.languages = []
        supported_languages = AppleOCREngine.languages(options)
        for lang in options.languages:
            if "+" in lang:
                raise ExitCodeException(
                    15, "Language combination with '+' is not supported by Apple OCR."
                )
            if lang not in supported_languages:
                raise ExitCodeException(
                    15,
                    f"Language '{lang}' is not supported by Apple OCR (supported in {options.appleocr_recognition_mode} mode: {', '.join(supported_languages)}). Use 'und' for undetermined language.",
                )

    # Need to populate this value, as OCRmyPDF core uses it to determine if OCR should be performed.
    # cf. https://github.com/ocrmypdf/OCRmyPDF/blob/main/src/ocrmypdf/_pipelines/ocr.py#L122
    options.tesseract_timeout = 1

    if options.pdf_renderer == "auto":
        options.pdf_renderer = "sandwich"


class AppleOCREngine(OcrEngine):
    """Implements OCR with Apple Vision Framework."""

    @staticmethod
    def version():
        return __version__

    @staticmethod
    def creator_tag(options):
        os_version = platform.mac_ver()[0]
        return f"AppleOCR Plugin {AppleOCREngine.version()} (on macOS {os_version})"

    def __str__(self):
        return f"AppleOCR Plugin {AppleOCREngine.version()}"

    @staticmethod
    def languages(options):
        if options.appleocr_recognition_mode == "livetext":
            return supported_languages_livetext
        elif options.appleocr_recognition_mode == "accurate":
            return supported_languages_accurate
        else:
            return supported_languages_fast

    @staticmethod
    def get_orientation(input_file, options):
        return tesseract.get_orientation(
            input_file,
            engine_mode=options.tesseract_oem,
            timeout=options.tesseract_non_ocr_timeout,
        )

    @staticmethod
    def get_deskew(input_file, options) -> float:
        return 0.0

    @staticmethod
    def generate_hocr(input_file, output_hocr, output_text, options):
        logging.debug("Starting OCR with Apple Vision Framework (hOCR renderer)...")

        ocr_result, width, height, _ = perform_ocr(Path(input_file), options)

        plaintext = "\n".join(tb.text for tb in ocr_result)

        hocr = build_hocr_document(ocr_result, width, height)
        with open(output_hocr, "w", encoding="utf-8") as f:
            f.write(hocr)
        with open(output_text, "w", encoding="utf-8") as f:
            f.write(plaintext)

    @staticmethod
    def generate_pdf(input_file, output_pdf, output_text, options):
        logging.debug("Starting OCR with Apple Vision Framework (sandwich renderer)...")

        (
            res,
            w,
            h,
            dpi,
        ) = perform_ocr(Path(input_file), options)
        plaintext = "\n".join(tb.text for tb in res)

        generate_pdf(dpi, w, h, 1.0, res, Path(output_pdf), True)

        with open(output_text, "w", encoding="utf-8") as f:
            f.write(plaintext)


@hookimpl
def get_ocr_engine():
    return AppleOCREngine()



# This code includes portions adapted from:
# https://github.com/ocrmypdf/OCRmyPDF-EasyOCR (MIT License)
# Copyright (c) 2023 James R. Barlow
# Modified by Masahiro Kiyota, 2025 to support vertical text in CJK languages

from __future__ import annotations

import importlib.resources
import math
from collections.abc import Iterable
from pathlib import Path

from pikepdf import (
    ContentStreamInstruction,
    Dictionary,
    Name,
    Operator,
    Pdf,
    unparse_content_stream,
)

from ocrmypdf_appleocr.common import Textbox, log

TEXT_POSITION_DEBUG = False
GLYPHLESS_FONT = importlib.resources.read_binary("ocrmypdf_appleocr", "pdf.ttf")
CHAR_ASPECT = 2
FONT_NAME = Name("/f-0-0")


def register_glyphlessfont(pdf: Pdf):
    """Register the glyphless font.

    Create several data structures in the Pdf to describe the font. While it create
    the data, a reference should be set in at least one page's /Resources dictionary
    to retain the font in the output PDF and ensure it is usable on that page.
    """
    PLACEHOLDER = Name.Placeholder

    basefont = pdf.make_indirect(
        Dictionary(
            BaseFont=Name.GlyphLessFont,
            DescendantFonts=[PLACEHOLDER],
            Encoding=Name("/Identity-H"),
            Subtype=Name.Type0,
            ToUnicode=PLACEHOLDER,
            Type=Name.Font,
        )
    )
    cid_font_type2 = pdf.make_indirect(
        Dictionary(
            BaseFont=Name.GlyphLessFont,
            CIDToGIDMap=PLACEHOLDER,
            CIDSystemInfo=Dictionary(
                Ordering="Identity",
                Registry="Adobe",
                Supplement=0,
            ),
            FontDescriptor=PLACEHOLDER,
            Subtype=Name.CIDFontType2,
            Type=Name.Font,
            DW=1000 // CHAR_ASPECT,
        )
    )
    basefont.DescendantFonts = [cid_font_type2]
    cid_font_type2.CIDToGIDMap = pdf.make_stream(b"\x00\x01" * 65536)
    basefont.ToUnicode = pdf.make_stream(
        b"/CIDInit /ProcSet findresource begin\n"
        b"12 dict begin\n"
        b"begincmap\n"
        b"/CIDSystemInfo\n"
        b"<<\n"
        b"  /Registry (Adobe)\n"
        b"  /Ordering (UCS)\n"
        b"  /Supplement 0\n"
        b">> def\n"
        b"/CMapName /Adobe-Identify-UCS def\n"
        b"/CMapType 2 def\n"
        b"1 begincodespacerange\n"
        b"<0000> <FFFF>\n"
        b"endcodespacerange\n"
        b"1 beginbfrange\n"
        b"<0000> <FFFF> <0000>\n"
        b"endbfrange\n"
        b"endcmap\n"
        b"CMapName currentdict /CMap defineresource pop\n"
        b"end\n"
        b"end\n"
    )
    font_descriptor = pdf.make_indirect(
        Dictionary(
            Ascent=1000,
            CapHeight=1000,
            Descent=-1,
            Flags=5,  # Fixed pitch and symbolic
            FontBBox=[0, 0, 1000 // CHAR_ASPECT, 1000],
            FontFile2=PLACEHOLDER,
            FontName=Name.GlyphLessFont,
            ItalicAngle=0,
            StemV=80,
            Type=Name.FontDescriptor,
        )
    )
    font_descriptor.FontFile2 = pdf.make_stream(GLYPHLESS_FONT)
    cid_font_type2.FontDescriptor = font_descriptor
    return basefont


class ContentStreamBuilder:
    def __init__(self, instructions=None):
        self._instructions = instructions or []

    def q(self):
        """Save the graphics state."""
        inst = [ContentStreamInstruction([], Operator("q"))]
        return ContentStreamBuilder(self._instructions + inst)

    def Q(self):
        """Restore the graphics state."""
        inst = [ContentStreamInstruction([], Operator("Q"))]
        return ContentStreamBuilder(self._instructions + inst)

    def cm(self, a: float, b: float, c: float, d: float, e: float, f: float):
        """Concatenate matrix."""
        inst = [ContentStreamInstruction([a, b, c, d, e, f], Operator("cm"))]
        return ContentStreamBuilder(self._instructions + inst)

    def BT(self):
        """Begin text object."""
        inst = [ContentStreamInstruction([], Operator("BT"))]
        return ContentStreamBuilder(self._instructions + inst)

    def ET(self):
        """End text object."""
        inst = [ContentStreamInstruction([], Operator("ET"))]
        return ContentStreamBuilder(self._instructions + inst)

    def BDC(self, mctype: Name, mcid: int):
        """Begin marked content sequence."""
        inst = [ContentStreamInstruction([mctype, Dictionary(MCID=mcid)], Operator("BDC"))]
        return ContentStreamBuilder(self._instructions + inst)

    def EMC(self):
        """End marked content sequence."""
        inst = [ContentStreamInstruction([], Operator("EMC"))]
        return ContentStreamBuilder(self._instructions + inst)

    def Tf(self, font: Name, size: int):
        """Set text font and size."""
        inst = [ContentStreamInstruction([font, size], Operator("Tf"))]
        return ContentStreamBuilder(self._instructions + inst)

    def Tm(self, a: float, b: float, c: float, d: float, e: float, f: float):
        """Set text matrix."""
        inst = [ContentStreamInstruction([a, b, c, d, e, f], Operator("Tm"))]
        return ContentStreamBuilder(self._instructions + inst)

    def Tr(self, mode: int):
        """Set text rendering mode."""
        inst = [ContentStreamInstruction([mode], Operator("Tr"))]
        return ContentStreamBuilder(self._instructions + inst)

    def Tz(self, scale: float):
        """Set text horizontal scaling."""
        inst = [ContentStreamInstruction([scale], Operator("Tz"))]
        return ContentStreamBuilder(self._instructions + inst)

    def TJ(self, text):
        """Show text."""
        inst = [ContentStreamInstruction([[text.encode("utf-16be")]], Operator("TJ"))]
        return ContentStreamBuilder(self._instructions + inst)

    def s(self):
        """Stroke and close path."""
        inst = [ContentStreamInstruction([], Operator("s"))]
        return ContentStreamBuilder(self._instructions + inst)

    def re(self, x: float, y: float, w: float, h: float):
        """Append rectangle to path."""
        inst = [ContentStreamInstruction([x, y, w, h], Operator("re"))]
        return ContentStreamBuilder(self._instructions + inst)

    def RG(self, r: float, g: float, b: float):
        """Set RGB stroke color."""
        inst = [ContentStreamInstruction([r, g, b], Operator("RG"))]
        return ContentStreamBuilder(self._instructions + inst)

    def w(self, width: float):
        """Set line width."""
        inst = [ContentStreamInstruction([width], Operator("w"))]
        return ContentStreamBuilder(self._instructions + inst)

    def S(self):
        """Stroke path."""
        inst = [ContentStreamInstruction([], Operator("S"))]
        return ContentStreamBuilder(self._instructions + inst)

    def m(self, x: float, y: float):
        """Move to point (x, y)."""
        inst = [ContentStreamInstruction([x, y], Operator("m"))]
        return ContentStreamBuilder(self._instructions + inst)

    def l(self, x: float, y: float):
        """Draw line to point (x, y)."""
        inst = [ContentStreamInstruction([x, y], Operator("l"))]
        return ContentStreamBuilder(self._instructions + inst)

    def h(self):
        """Close path."""
        inst = [ContentStreamInstruction([], Operator("h"))]
        return ContentStreamBuilder(self._instructions + inst)

    def build(self):
        return self._instructions

    def add(self, other: ContentStreamBuilder):
        return ContentStreamBuilder(self._instructions + other._instructions)


def generate_text_content_stream(
    results: Iterable[Textbox],
    scale: tuple[float, float],
    height: int,
    boxes=False,
):
    """Generate a content stream for the described by results.

    Args:
        results (Iterable[Textbox]): Results of OCR.
        scale (tuple[float, float]): Scale of the image.
        height (int): Height of the image.

    Yields:
        ContentStreamInstruction: Content stream instructions.
    """

    cs = ContentStreamBuilder()
    cs = cs.add(cs.q())
    for n, result in enumerate(results):
        text = result.text
        # bbox is in up-left origin, pixel coordinates
        bbox = result.bb
        box_width = bbox.true_width() * scale[0]
        box_height = bbox.true_height() * scale[1]
        angle = -bbox.angle()  # PDF coordinate is y-up, so negate the angle
        ulx = bbox.ul.x * scale[0]
        uly = (height - bbox.ul.y) * scale[1]
        llx = bbox.ll.x * scale[0]
        lly = (height - bbox.ll.y) * scale[1]
        cos_a = math.cos(angle)
        sin_a = math.sin(angle)

        if len(text) == 0 or box_width <= 0 or box_height <= 0:
            continue

        vertical = result.is_vertical

        log.debug(
            f"Textline '{text}' bbox (in px): {bbox} vertical: {vertical}, angle: {angle}, box_width: {box_width}, box_height: {box_height}"
        )

        if vertical:
            font_size = box_width
            stretch = 100.0 * box_height / len(text) / font_size * CHAR_ASPECT
            tm_args = (
                font_size * sin_a,
                -font_size * cos_a,
                font_size * cos_a,
                -font_size * sin_a,
                ulx,
                uly,
            )
        else:
            font_size = box_height
            stretch = 100.0 * box_width / len(text) / font_size * CHAR_ASPECT
            tm_args = (
                font_size * cos_a,
                font_size * sin_a,
                -font_size * sin_a,
                font_size * cos_a,
                llx,
                lly,
            )

        cs = cs.add(
            ContentStreamBuilder()
            .BT()
            .BDC(Name.Span, n)
            .Tr(3)  # Invisible ink
            .Tm(*tm_args)
            .Tf(FONT_NAME, 1)
            .Tz(stretch)
            .TJ(text)
            .EMC()
            .ET()
        )
        if boxes:
            urx = bbox.ur.x * scale[0]
            ury = (height - bbox.ur.y) * scale[1]
            lrx = bbox.lr.x * scale[0]
            lry = (height - bbox.lr.y) * scale[1]
            cs = cs.add(
                ContentStreamBuilder()
                .q()
                .RG(1.0, 0.0, 0.0)
                .w(0.75)
                .m(ulx, uly)
                .l(urx, ury)
                .l(lrx, lry)
                .l(llx, lly)
                .h()
                .S()
                .Q()
            )

    cs = cs.Q()
    return cs.build()


def generate_pdf(
    dpi: tuple[float, float],
    width: int,
    height: int,
    image_scale: float,
    results: Iterable[Textbox],
    output_pdf: Path,
    boxes: bool,
):
    """Convert OCR results to a PDF with text annotations (no images).

    Args:
        dpi: DPI of the OCR image.
        width: Width of the OCR image in pixels.
        height: Height of the OCR image in pixels.
        image_scale: Scale factor applied to the OCR image. 1.0 means the
            image is at the scale implied by its DPI. 2.0 means the image
            is twice as large as implied by its DPI.
        results: List of Textbox objects.
        output_pdf: Path to the output PDF file that this will function will
            create.

    Returns:
        output_pdf
    """

    scale = 72.0 / dpi[0] / image_scale, 72.0 / dpi[1] / image_scale

    with Pdf.new() as pdf:
        pdf.add_blank_page(page_size=(width * scale[0], height * scale[1]))
        pdf.pages[0].Resources = Dictionary(
            Font=Dictionary(
                {
                    str(FONT_NAME): register_glyphlessfont(pdf),
                }
            )
        )

        cs = generate_text_content_stream(results, scale, height, boxes=boxes)
        pdf.pages[0].Contents = pdf.make_stream(unparse_content_stream(cs))

        pdf.save(output_pdf)
    return output_pdf



import math
import multiprocessing as mp
import platform
from pathlib import Path

import Cocoa
import objc
import Vision
from PyObjCTools import AppHelper

Vision  # to silence unused import warning

from ocrmypdf_appleocr.common import BoundingBox, Point, Textbox, locale_to_lang_code, log

livetext_supported = int(platform.mac_ver()[0].split(".")[0]) >= 13  # macOS Ventura or later
supported_languages_livetext: list[str] = []
if livetext_supported:
    app_info = Cocoa.NSBundle.mainBundle().infoDictionary()
    app_info["LSBackgroundOnly"] = "1"
    objc.registerMetaDataForSelector(
        b"VKCImageAnalyzer",
        b"processRequest:progressHandler:completionHandler:",
        {
            "arguments": {
                3: {
                    "callable": {
                        "retval": {"type": b"v"},
                        "arguments": {
                            0: {"type": b"^v"},
                            1: {"type": b"d"},
                        },
                    }
                },
                4: {
                    "callable": {
                        "retval": {"type": b"v"},
                        "arguments": {
                            0: {"type": b"^v"},
                            1: {"type": b"@"},
                            2: {"type": b"@"},
                        },
                    }
                },
            }
        },
    )
    VKCImageAnalyzer = objc.lookUpClass("VKCImageAnalyzer")
    VKCImageAnalyzerRequest = objc.lookUpClass("VKCImageAnalyzerRequest")
    with objc.autorelease_pool():
        for locale in VKCImageAnalyzer.supportedRecognitionLanguages():
            if locale in locale_to_lang_code:
                supported_languages_livetext.append(locale_to_lang_code[locale])
            else:
                log.debug(f"Locale '{locale}' not mapped to any language code.")


def _ocr_VKCImageAnalyzerRequest_child_main(
    q: mp.Queue, image_file: Path, width: int, height: int, locales: list[str]
):
    result = None

    def _process_handler(analysis, error):
        nonlocal result
        lines: list[Textbox] = []
        response_lines = analysis.allLines()
        if response_lines:
            for l in response_lines:
                pts = []
                quad = l.quad()
                pts.append((quad.topLeft().x * width, quad.topLeft().y * height))
                pts.append((quad.topRight().x * width, quad.topRight().y * height))
                pts.append((quad.bottomRight().x * width, quad.bottomRight().y * height))
                pts.append((quad.bottomLeft().x * width, quad.bottomLeft().y * height))
                cx = sum(p[0] for p in pts) / 4.0
                cy = sum(p[1] for p in pts) / 4.0

                # sort points in CW order in x-right, y-down coordinate system
                pts.sort(key=lambda p: math.atan2(p[1] - cy, p[0] - cx))

                ideal_ul_angle = -3 * math.pi / 4
                min_diff = float("inf")
                ul_index = -1
                for i, p in enumerate(pts):
                    angle = math.atan2(p[1] - cy, p[0] - cx)
                    diff = abs(angle - ideal_ul_angle)
                    if diff > math.pi:
                        diff = abs(diff - 2 * math.pi)
                    if diff < min_diff:
                        min_diff = diff
                        ul_index = i
                ul = Point(int(pts[ul_index][0]), int(pts[ul_index][1]))
                ur = Point(int(pts[(ul_index + 1) % 4][0]), int(pts[(ul_index + 1) % 4][1]))
                lr = Point(int(pts[(ul_index + 2) % 4][0]), int(pts[(ul_index + 2) % 4][1]))
                ll = Point(int(pts[(ul_index + 3) % 4][0]), int(pts[(ul_index + 3) % 4][1]))
                bb = BoundingBox(ul, ur, ll, lr)
                # https://github.com/WebKit/WebKit/blob/main/Source/WebKit/Platform/cocoa/ImageAnalysisUtilities.mm
                is_vert = l.layoutDirection() == 5
                lines.append(
                    Textbox(
                        l.string(),
                        bb,
                        100,  # VKCImageAnalyzer does not provide confidence score,
                        is_vert,
                    )
                )
        AppHelper.stopEventLoop()
        result = lines

    with objc.autorelease_pool():
        analyzer = VKCImageAnalyzer.alloc().init()
        nsimg = Cocoa.NSImage.alloc().initWithContentsOfFile_(image_file.as_posix())
        req = VKCImageAnalyzerRequest.alloc().initWithImage_requestType_(
            nsimg,
            1,  # 1 == VKAnalysisTypeText
        )
        req.setLocales_(locales)
        analyzer.processRequest_progressHandler_completionHandler_(
            req, lambda progress: None, _process_handler
        )
        # Wait for completion
        AppHelper.runConsoleEventLoop()

    if result is None:
        result = []
        raise RuntimeError("Error in performing VKCImageAnalyzerRequest")

    q.put(result)


def ocr_VKCImageAnalyzerRequest(
    image_file: Path, width: int, height: int, locales: list[str]
) -> list[Textbox]:
    result = None

    mp.set_start_method("spawn", force=True)
    q = mp.Queue()
    p = mp.Process(
        target=_ocr_VKCImageAnalyzerRequest_child_main,
        args=(q, image_file, width, height, locales),
    )
    p.start()
    p.join(30)  # Timeout after 30 seconds
    if p.is_alive():
        p.terminate()
        p.join()
        raise RuntimeError("Timeout in VKCImageAnalyzerRequest")
    if not q.empty():
        result = q.get()
    if result is None:
        result = []
        raise RuntimeError("Error in performing VKCImageAnalyzerRequest")

    return result



from pathlib import Path

import Cocoa
import objc
import Vision

from ocrmypdf_appleocr.common import (
    BoundingBox,
    Point,
    Textbox,
    lang_code_to_locale,
    locale_to_lang_code,
    log,
)

supported_languages_accurate: list[str] = []
supported_languages_fast: list[str] = []
with objc.autorelease_pool():
    req = Vision.VNRecognizeTextRequest.alloc().init()
    req.setRecognitionLevel_(Vision.VNRequestTextRecognitionLevelAccurate)
    lst, _ = req.supportedRecognitionLanguagesAndReturnError_(None)
    for locale in lst:
        if locale in locale_to_lang_code:
            supported_languages_accurate.append(locale_to_lang_code[locale])
        else:
            log.debug(f"Locale '{locale}' not mapped to any language code.")
    req.setRecognitionLevel_(Vision.VNRequestTextRecognitionLevelFast)
    lst, _ = req.supportedRecognitionLanguagesAndReturnError_(None)
    for locale in lst:
        if locale in locale_to_lang_code:
            supported_languages_fast.append(locale_to_lang_code[locale])
        else:
            log.debug(f"Locale '{locale}' not mapped to any language code.")


def ocr_VNRecognizeTextRequest(image_file: Path, width: int, height: int, options) -> list[Textbox]:
    with objc.autorelease_pool():
        recognize_request = Vision.VNRecognizeTextRequest.alloc().init()
        if options.languages:
            locales = [lang_code_to_locale.get(lang, lang) for lang in options.languages]
            recognize_request.setAutomaticallyDetectsLanguage_(False)
            recognize_request.setRecognitionLanguages_(locales)
        else:
            log.debug("Using automatic language detection.")
            recognize_request.setAutomaticallyDetectsLanguage_(True)
        if options.appleocr_disable_correction:
            recognize_request.setUsesLanguageCorrection_(False)
        else:
            recognize_request.setUsesLanguageCorrection_(True)
        level = options.appleocr_recognition_mode
        if level == "fast":
            recognize_request.setRecognitionLevel_(Vision.VNRequestTextRecognitionLevelFast)
        elif level == "accurate":
            recognize_request.setRecognitionLevel_(Vision.VNRequestTextRecognitionLevelAccurate)
        else:
            raise RuntimeError(
                f"VNRecognizeTextRequest does not support recognition level '{level}'"
            )
        request_handler = Vision.VNImageRequestHandler.alloc().initWithURL_options_(
            Cocoa.NSURL.fileURLWithPath_(image_file.absolute().as_posix()), None
        )
        _, error = request_handler.performRequests_error_([recognize_request], None)
        if error:
            raise RuntimeError(f"Error in performing VNRecognizeTextRequest: {error=}")
        res: list[Textbox] = []
        for o in recognize_request.results():
            recognized_text: Vision.VNRecognizedText = o.topCandidates_(1)[0]
            bb = o.boundingBox()  # bb in bottom-left origin, normalized coordinates
            b = BoundingBox(  # b in upper-left origin, pixel coordinates
                Point(
                    int(bb.origin.x * width), int((1 - bb.origin.y - bb.size.height) * height)
                ),  # ul
                Point(
                    int((bb.origin.x + bb.size.width) * width),
                    int((1 - bb.origin.y - bb.size.height) * height),
                ),  # ur
                Point(int(bb.origin.x * width), int((1 - bb.origin.y) * height)),  # ll
                Point(
                    int((bb.origin.x + bb.size.width) * width), int((1 - bb.origin.y) * height)
                ),  # lr
            )
            confidence = recognized_text.confidence()
            text = recognized_text.string()
            res.append(
                Textbox(
                    text,
                    b,
                    int(confidence * 100),
                    False,  # VNRecognizeTextRequest does not provide orientation info
                )
            )
    return res



from ocrmypdf_appleocr.common import Textbox


def build_hocr_line(textbox: Textbox, page_number: int, line_number: int, lang: str) -> str:
    text, bb, confidence, is_vert = textbox
    if not is_vert:
        bbox = f"{bb.to_hocr_bbox()}; {bb.estimated_baseline()}"
    else:
        bbox = f"{bb.to_hocr_bbox()}"
    text = (
        text.replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
        .replace("'", "&apos;")
    )
    return f"""<div class="ocr_carea" id="block_{page_number}_{line_number}" title="{bbox}">
<p class="ocr_par" id="par_{page_number}_{line_number}" lang="{lang}" title="{bbox}">
  <span class="ocr_line" id="line_{page_number}_{line_number}" title="{bbox}">
    <span class="ocrx_word" id="word_{page_number}_{line_number}" title="{bbox}; x_wconf {confidence}">{text}</span>
  </span>
</p>
</div>
"""


hocr_template = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<title></title>
<meta http-equiv="Content-Type" content="text/html;charset=utf-8"/>
<meta name="ocr-system" content="" />
<meta name="ocr-capabilities" content="ocr_page ocr_carea ocr_par ocr_line ocrx_word"/>
</head>
<body>
<div class="ocr_page" id="page_0" title="bbox 0 0 {width} {height}">
{content}
</div>
</body>
</html>
"""


def build_hocr_document(ocr_result: list[Textbox], width, height) -> str:
    content = "".join(build_hocr_line(tb, 0, i, "und") for i, tb in enumerate(ocr_result))
    return hocr_template.format(content=content, width=width, height=height)



import logging
import math
from typing import NamedTuple

log = logging.getLogger(__name__)


class Point(NamedTuple):
    """Represents a point in 2D space. (upper-left origin, pixel coordinates)"""

    x: int
    y: int


def distance(p1: Point, p2: Point) -> float:
    return math.sqrt((p1.x - p2.x) ** 2 + (p1.y - p2.y) ** 2)


class BoundingBox(NamedTuple):
    """Represents a bounding box in the document."""

    ul: Point
    ur: Point
    ll: Point
    lr: Point

    def to_hocr_bbox(self) -> str:
        return f"bbox {self.ul.x} {self.ul.y} {self.lr.x} {self.lr.y}"

    def estimated_baseline(self) -> str:
        intercept = max(self.ll.y, self.lr.y) - self.lr.y
        slope = (
            (self.lr.y - self.ll.y) / (self.lr.x - self.ll.x) if (self.lr.x - self.ll.x) != 0 else 0
        )
        return f"baseline {slope} {intercept}"

    def true_width(self) -> float:
        """Returns the true width of the bounding box."""
        return (distance(self.ul, self.ur) + distance(self.ll, self.lr)) / 2.0

    def true_height(self) -> float:
        """Returns the true height of the bounding box."""
        return (distance(self.ul, self.ll) + distance(self.ur, self.lr)) / 2.0

    def angle(self) -> float:
        """Returns the angle of the bounding box in radians."""
        delta_x = self.ur.x - self.ul.x
        delta_y = self.ur.y - self.ul.y
        return math.atan2(delta_y, delta_x)


class Textbox(NamedTuple):
    """Represents a text box in the document."""

    text: str
    bb: BoundingBox
    confidence: int  # 0-100
    is_vertical: bool


lang_code_to_locale = {
    "eng": "en-US",
    "fra": "fr-FR",
    "ita": "it-IT",
    "deu": "de-DE",
    "spa": "es-ES",
    "por": "pt-BR",
    "chi_sim": "zh-Hans",
    "chi_tra": "zh-Hant",
    "yue_sim": "yue-Hans",
    "yue_tra": "yue-Hant",
    "kor": "ko-KR",
    "jpn": "ja-JP",
    "rus": "ru-RU",
    "ukr": "uk-UA",
    "tha": "th-TH",
    "vie": "vi-VT",
    "ara": "ar-SA",
    "ars": "ars-SA",
    "tur": "tr-TR",
    "ind": "id-ID",
    "ces": "cs-CZ",
    "dan": "da-DK",
    "nld": "nl-NL",
    "nor": "no-NO",
    "nno": "nn-NO",
    "nob": "nb-NO",
    "msa": "ms-MY",
    "pol": "pl-PL",
    "ron": "ro-RO",
    "swe": "sv-SE",
}

locale_to_lang_code = {v: k for k, v in lang_code_to_locale.items()}
